"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.mapNodeList = mapNodeList;
exports.bindChildren = bindChildren;
exports.translateNode = translateNode;
exports.translateDocument = translateDocument;

var _TText = require("../tree/TText");

var _TPhrasing = require("../tree/TPhrasing");

var _TBlock = require("../tree/TBlock");

var _TDocument = require("../tree/TDocument");

var _toSerializable = require("../dom/to-serializable");

var _TEmpty = require("../tree/TEmpty");

var _defaultHTMLElementModels = _interopRequireDefault(require("../model/defaultHTMLElementModels"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function mapNodeList({
  nodeList,
  parent,
  parentStyles,
  params
}) {
  const nextMap = [];

  for (const i in nodeList) {
    const child = nodeList[i];
    const translated = translateNode({
      node: child,
      parentStyles,
      params,
      nodeIndex: Number(i),
      parent
    });

    if (translated) {
      nextMap.push(translated);
    }
  }

  return nextMap;
}

function bindChildren(node, children, params) {
  if (!node.elementModel || !node.elementModel.isOpaque) {
    node.bindChildren(mapNodeList({
      nodeList: children,
      parent: node,
      parentStyles: node.styles,
      params
    }));
  }
}

function translateElement({
  node,
  nodeIndex,
  params,
  parent,
  parentStyles
}) {
  const tagName = node.tagName.toLowerCase();
  const sharedProps = {
    tagName,
    nodeIndex,
    parentStyles,
    stylesMerger: params.stylesMerger,
    attributes: node.attribs,
    domNode: null,
    parent
  };

  if (tagName === 'html') {
    const tdoc = new _TDocument.TDocument({ ...sharedProps,
      styles: parentStyles
    });
    bindChildren(tdoc, node.children, params);
    tdoc.parseChildren();
    return tdoc;
  }

  const elementModel = params.modelRegistry.getElementModelFromTagName(tagName);

  if (!elementModel) {
    return new _TEmpty.TEmpty({ ...sharedProps,
      isUnregistered: true,
      contentModel: null,
      elementModel: null,
      domNode: node,
      nodeIndex: 0
    });
  }

  const contentModel = elementModel.contentModel;

  if (elementModel.isTranslatableTextual()) {
    if (node.children.length === 1) {
      const child = node.children[0];

      if ((0, _toSerializable.isSerializableText)(child)) {
        return new _TText.TText({ ...sharedProps,
          contentModel,
          elementModel,
          data: child.data
        });
      }
    } else if (node.children.length === 0) {
      return new _TText.TText({ ...sharedProps,
        contentModel,
        elementModel,
        data: ''
      });
    }

    const phrasing = new _TPhrasing.TPhrasing({ ...sharedProps,
      contentModel,
      elementModel
    });
    bindChildren(phrasing, node.children, params);
    return phrasing;
  }

  if (elementModel.isTranslatableBlock()) {
    const block = new _TBlock.TBlock({ ...sharedProps,
      contentModel,
      elementModel,
      parentStyles,
      domNode: elementModel.isOpaque ? node : null
    });
    bindChildren(block, node.children, params);
    return block;
  }

  return new _TEmpty.TEmpty({ ...sharedProps,
    isUnregistered: false,
    contentModel,
    elementModel,
    domNode: node
  });
}

function translateNode({
  node,
  parentStyles,
  params,
  nodeIndex,
  parent
}) {
  if ((0, _toSerializable.isSerializableText)(node)) {
    return new _TText.TText({
      data: node.data,
      stylesMerger: params.stylesMerger,
      contentModel: null,
      elementModel: null,
      parentStyles,
      domNode: null,
      nodeIndex,
      parent
    });
  }

  if ((0, _toSerializable.isSerializableElement)(node)) {
    return translateElement({
      node,
      parentStyles,
      params,
      nodeIndex,
      parent
    });
  }

  return null;
}

function translateDocument(documentTree, params) {
  const serializableDocTree = (0, _toSerializable.toSerializableChildren)(documentTree);
  const rootNodes = mapNodeList({
    nodeList: serializableDocTree,
    parentStyles: params.baseStyles,
    parent: null,
    params
  });
  let foundTdoc = rootNodes.find(n => n instanceof _TDocument.TDocument);

  if (foundTdoc) {
    return foundTdoc;
  } else {
    let body = rootNodes.find(n => n.tagName === 'body');
    const newTdoc = new _TDocument.TDocument({
      styles: params.baseStyles,
      stylesMerger: params.stylesMerger,
      domNode: null,
      nodeIndex: 0,
      parent: null
    });

    if (!body) {
      body = new _TBlock.TBlock({
        tagName: 'body',
        stylesMerger: params.stylesMerger,
        contentModel: _defaultHTMLElementModels.default.body.contentModel,
        elementModel: _defaultHTMLElementModels.default.body,
        parentStyles: params.baseStyles,
        domNode: null,
        nodeIndex: 0,
        parent: newTdoc
      });
      body.bindChildren(rootNodes);
    }

    newTdoc.bindChildren([body]);
    return newTdoc;
  }
}
//# sourceMappingURL=translate.js.map